# Copyright Red Hat
#
# snapm/manager/__init__.py - Snapshot Manager
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: Apache-2.0
"""
Top level interface to the snapshot manager.
"""

from ._manager import *  # noqa: F401, F403
from ._manager import __all__  # noqa: F401
